from enum import Enum


class CardNetwork(str, Enum):
    atm = 'atm'
    visa = 'visa'


class CardStatus(str, Enum):
    active = 'active'
    blocked = 'blocked'
    created = 'created'
    deactivated = 'deactivated'
    printing = 'printing'


class CardTransactionType(str, Enum):
    auth = 'auth'
    capture = 'capture'
    expiration = 'expiration'
    refund = 'refund'
    void = 'void'


class CardType(str, Enum):
    physical = 'physical'
    virtual = 'virtual'


class DepositNetwork(str, Enum):
    cash = 'cash'
    internal = 'internal'
    spei = 'spei'


class ServiceProviderCategory(str, Enum):
    cable = 'cable'
    credit_card = 'credit_card'
    electricity = 'electricity'
    gas = 'gas'
    internet = 'internet'
    landline_telephone = 'landline_telephone'
    mobile_telephone_postpaid = 'mobile_telephone_postpaid'
    mobile_telephone_prepaid = 'mobile_telephone_prepaid'
    satelite_televesion = 'satelite_televesion'
    water = 'water'


class TransactionStatus(str, Enum):
    created = 'created'
    submitted = 'submitted'
    in_review = 'in_review'
    succeeded = 'succeeded'
    failed = 'failed'


class TransferNetwork(str, Enum):
    internal = 'internal'
    spei = 'spei'


class CommissionType(str, Enum):
    card_request = 'card_request'
    cash_deposit = 'cash_deposit'


class EntryType(str, Enum):
    credit = 'credit'
    debit = 'debit'
    commission = 'commission'



import datetime as dt
import json
from dataclasses import dataclass
from enum import Enum

import pytest
from pydantic import BaseModel

from cuenca_validations.types import (
    CardQuery,
    JSONEncoder,
    QueryParams,
    SantizedDict,
    TransactionStatus,
    digits,
)

today = dt.date.today()
now = dt.datetime.now()
utcnow = now.astimezone(dt.timezone.utc)


class TestEnum(Enum):
    zero = 0


@dataclass
class TestClass:
    uno: str

    def to_dict(self):
        return dict(uno=self.uno, dos='dos')


def test_dict():
    model = QueryParams(count=1, created_before=now)
    assert model.dict() == dict(count=1, created_before=utcnow.isoformat())


def test_dict_with_exclude():
    model = QueryParams(count=1, created_before=now, user_id='USXXXX')
    assert model.dict(exclude={'user_id'}) == dict(
        count=1, created_before=utcnow.isoformat()
    )


def test_dict_with_exclude_unset():
    model = QueryParams(count=1, created_before=now)
    assert model.dict(exclude_unset=False) == dict(
        count=1, created_before=utcnow.isoformat(), page_size=100
    )


def test_sanitized_dict():
    assert SantizedDict(
        status=TransactionStatus.succeeded, time=now, hello='there'
    ) == dict(status='succeeded', time=utcnow.isoformat(), hello='there')


@pytest.mark.parametrize(
    'count, truth',
    [
        (1, True),
        ('1', True),
        (True, True),
        (False, False),
        (0, False),
        ('0', False),
    ],
)
def test_count(count, truth):
    q = QueryParams(count=count)
    assert q.count is truth


@pytest.mark.parametrize(
    'value, result',
    [
        (TestEnum.zero, 0),
        (today, today.isoformat()),
        (now, utcnow.isoformat()),
        (TestClass(uno='uno'), dict(uno='uno', dos='dos')),
    ],
)
def test_json_encoder(value, result):
    to_encode = dict(value=value)
    encoded = json.dumps(to_encode, cls=JSONEncoder)
    decoded = json.loads(encoded)

    assert decoded['value'] == result


def test_invalid_class():
    """
    For a class that doesn't have a `to_dict` method and it is not a type of
    `date` nor `Enum`, will use the default `json.JSONEncoder` method which
    raises a `TypeError`.
    """

    class ClassWithoutToDict:
        ...

    invalid_class = ClassWithoutToDict()
    with pytest.raises(TypeError):
        json.dumps(invalid_class, cls=JSONEncoder)


class Accounts(BaseModel):
    number: digits(5, 8)  # type: ignore


def test_only_digits():
    acc = Accounts(number='123456')
    assert acc.number == '123456'


@pytest.mark.parametrize(
    'number, error',
    [
        ('123', 'value_error.any_str.min_length'),
        ('1234567890', 'value_error.any_str.max_length'),
        ('no_123', 'value_error.digits'),
    ],
)
def test_invalid_digits(number, error):
    with pytest.raises(ValueError) as exception:
        Accounts(number=number)
    assert error in str(exception)


def test_card_query_exp_cvv_if_number_set():
    values = dict(number='123456', exp_month=1, exp_year=2026)
    card_query = CardQuery(**values)
    assert all(
        getattr(card_query, key) == value for key, value in values.items()
    )


@pytest.mark.parametrize(
    'input_value',
    [
        (dict(exp_month=1)),
        (dict(exp_year=2026)),
        (dict(cvv2='123')),
        (dict(cvv='123')),
    ],
)
def test_card_query_exp_cvv_if_number_not_set(input_value):
    with pytest.raises(ValueError):
        CardQuery(**input_value)



import re

from enums import EntryType
from pydantic import BaseModel, validator

mapper = dict(
    credit={
        'Deposit': ['TR', 'SP', 'LT', 'CD'],
        'CardTransaction': ['CT'],
    },
    debit={
        'BillPayment': ['ST'],
        'Transfer': ['TR', 'SP', 'LT'],
        'WhatsappTransfer': ['SW'],
        'Commission': ['CO'],
        'CardTransaction': ['CT'],
    },
    commission={
        'Deposit': ['TR', 'SP', 'LT', 'CD'],
        'Transfer': ['TR', 'SP', 'LT'],
    },
)


class RelatedTransaction(str):
    def __init__(cls, uri: str):
        cls.uri = uri
        cls.id = cls._get_id(uri)
        cls.mapper_ids = cls._mapper_ids()

    @classmethod
    def __get_validators__(cls):
        yield cls
        yield cls.validate

    @classmethod
    def validate(cls, tr: 'Transaction') -> 'CallableGenerator':
        if tr.id[:2] not in tr.mapper_ids:
            raise ValueError('invalid id format')
        return tr

    @staticmethod
    def _get_id(uri: str) -> str:
        match = re.search(r'/[a-z]+/(\w+)', uri)
        return match.group(1)

    @staticmethod
    def _mapper_ids():
        return list(set(re.findall(r"'([A-Z]{0,2})'", str(mapper))))

    def get_model(cls, entry_type) -> str:
        if type(entry_type) is not EntryType:
            raise ValueError('The required enum is EntryType')
        return next(
            (
                model
                for model, types in mapper[entry_type.value].items()
                if cls.id[:2] in types
            ),
            None,
        )



from typing import Optional, Union

from clabe import Clabe
from pydantic import BaseModel, Extra, StrictStr

from ..types.enums import CardStatus
from .card import PaymentCardNumber, StrictPaymentCardNumber
from .general import StrictPositiveInt


class TransferRequest(BaseModel):
    recipient_name: StrictStr
    account_number: Union[Clabe, PaymentCardNumber]
    amount: StrictPositiveInt  # in centavos
    descriptor: StrictStr  # how it'll appear for the recipient
    idempotency_key: str  # must be unique for each transfer


class StrictTransferRequest(TransferRequest):
    account_number: Union[Clabe, StrictPaymentCardNumber]


class CardUpdateRequest(BaseModel):
    user_id: Optional[str]
    ledger_account_id: Optional[str]
    status: Optional[CardStatus]

    class Config:
        extra = Extra.forbid


class CardRequest(BaseModel):
    user_id: str
    ledger_account_id: str



__all__ = [
    'ApiKeyQuery',
    'BillPaymentQuery',
    'CardNetwork',
    'CardQuery',
    'CardStatus',
    'CardTransactionType',
    'CardType',
    'CommissionType',
    'DepositNetwork',
    'DepositQuery',
    'JSONEncoder',
    'EntryType',
    'PageSize',
    'PaymentCardNumber',
    'QueryParams',
    'SantizedDict',
    'ServiceProviderCategory',
    'StrictPaymentCardNumber',
    'StrictPositiveInt',
    'StrictPositiveFloat',
    'StrictTransferRequest',
    'RelatedTransaction',
    'TransactionQuery',
    'TransactionStatus',
    'TransferNetwork',
    'TransferQuery',
    'TransferRequest',
    'digits',
]

from .card import PaymentCardNumber, StrictPaymentCardNumber
from .transaction import RelatedTransaction
from .enums import (
    CardNetwork,
    CardStatus,
    CardTransactionType,
    CardType,
    CommissionType,
    DepositNetwork,
    ServiceProviderCategory,
    TransactionStatus,
    TransferNetwork,
)
from .general import (
    JSONEncoder,
    SantizedDict,
    StrictPositiveFloat,
    StrictPositiveInt,
    digits,
)
from .queries import (
    ApiKeyQuery,
    BillPaymentQuery,
    CardQuery,
    DepositQuery,
    PageSize,
    QueryParams,
    TransactionQuery,
    TransferQuery,
)
from .requests import StrictTransferRequest, TransferRequest



import pytest
from pydantic import ValidationError

from cuenca_validations.types import EntryModel


def test_entry_model():
    entry = EntryModel(id='SPXXX', type='credit')
    assert entry.get_model() == 'Deposit'


def test_entry_model_none_for_type_incorrect():
    entry = EntryModel(id='CDXXX', type='debit')
    assert not entry.get_model()


def test_invalid_id_entry_model():
    with pytest.raises(ValidationError) as exc_info:
        EntryModel(id='XXXXX', type='credit')
    assert exc_info.value.errors()[0] == dict(
        loc=('id',),
        type='value_error',
        msg='invalid id format',
    )


def test_invalid_type_entry_model():
    with pytest.raises(ValidationError) as exc_info:
        EntryModel(id='SPXXX', type='foo')
    assert exc_info.value.errors()[0] == dict(
        loc=('type',),
        type='value_error',
        msg='invalid entry_type format',
    )



from importlib.machinery import SourceFileLoader

from setuptools import find_packages, setup

version = SourceFileLoader(
    'version', 'cuenca_validations/version.py'
).load_module()


with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='cuenca_validations',
    version=version.__version__,
    author='Cuenca',
    author_email='dev@cuenca.com',
    description='Cuenca common validations',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cuenca-mx/cuenca-validations',
    packages=find_packages(),
    include_package_data=True,
    package_data=dict(cuenca_validations=['py.typed']),
    python_requires='>=3.6',
    install_requires=[
        'clabe>=1.2,<1.3',
        'pydantic>=1.6,<1.7',
        'dataclasses>=0.6;python_version<"3.7"',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



from typing import TYPE_CHECKING, ClassVar

from pydantic.types import PaymentCardNumber as PydanticPaymentCardNumber
from pydantic.validators import (
    constr_length_validator,
    constr_strip_whitespace,
    str_validator,
)

from ..card_bins import CARD_BINS
from ..errors import CardBinValidationError

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator


class PaymentCardNumber(PydanticPaymentCardNumber):
    min_length: ClassVar[int] = 16
    max_length: ClassVar[int] = 16

    def __init__(self, card_number: str):
        self.bin = card_number[:6]
        self.last4 = card_number[-4:]
        self.brand = self._get_brand(card_number)
        self.bank_code = CARD_BINS.get(self.bin)

    @classmethod
    def __get_validators__(cls) -> 'CallableGenerator':
        yield str_validator
        yield constr_strip_whitespace
        yield constr_length_validator
        yield cls.validate_digits
        yield cls.validate_luhn_check_digit
        yield cls


class StrictPaymentCardNumber(PaymentCardNumber):
    """
    requires that the BIN be associated to a known BIN for a Mexican bank
    """

    bank_code: str

    @classmethod
    def __get_validators__(cls) -> 'CallableGenerator':
        yield from PaymentCardNumber.__get_validators__()
        yield cls.validate_bin

    @classmethod
    def validate_bin(cls, card_number: PaymentCardNumber):
        if card_number.bank_code is None:
            raise CardBinValidationError
        return card_number



__all__ = ['CardBinValidationError', 'NotDigitError']

from pydantic.errors import (
    NotDigitError as PydanticNotDigitError,
    PydanticValueError,
)


class CardBinValidationError(PydanticValueError):
    code = 'payment_card_number.bin'
    msg_template = (
        'The card number contains a BIN (first six digits) that does not have'
        'a known association with a Mexican bank. To add the association,'
        'please file an issue:'
        'https://github.com/cuenca-mx/cuenca-validations/issues'
    )


class NotDigitError(PydanticNotDigitError):
    code = 'digits'
    msg_template = 'value is not all digits'
